#!/usr/bin/perl
#!/usr/bin/perl -d:ptkdb

###############################################################################
#
# libmap.pl -  Conversion GPS and Map-X/Y Coordinates
#
# Copyright (C) 2009  Rainer Walther  (rainerwalther-mail@web.de)
#
# Creative Commons Lizenz mit den Zusaetzen (by, nc, sa)
#
# Es ist Ihnen gestattet: 
#     * das Werk vervielfltigen, verbreiten und ffentlich zugnglich machen
#     * Abwandlungen bzw. Bearbeitungen des Inhaltes anfertigen
# 
# Zu den folgenden Bedingungen:
#     * Namensnennung.
#       Sie mssen den Namen des Autors/Rechteinhabers in der von ihm festgelegten Weise nennen.
#     * Keine kommerzielle Nutzung.
#       Dieses Werk darf nicht fr kommerzielle Zwecke verwendet werden.
#     * Weitergabe unter gleichen Bedingungen.
#       Wenn Sie den lizenzierten Inhalt bearbeiten oder in anderer Weise umgestalten,
#       verndern oder als Grundlage fr einen anderen Inhalt verwenden,
#       drfen Sie den neu entstandenen Inhalt nur unter Verwendung von Lizenzbedingungen
#       weitergeben, die mit denen dieses Lizenzvertrages identisch oder vergleichbar sind.
# 
# Im Falle einer Verbreitung mssen Sie anderen die Lizenzbedingungen, unter welche dieses
# Werk fllt, mitteilen. Am Einfachsten ist es, einen Link auf diese Seite einzubinden.
# 
# Jede der vorgenannten Bedingungen kann aufgehoben werden, sofern Sie die Einwilligung
# des Rechteinhabers dazu erhalten.
# 
# Diese Lizenz lsst die Urheberpersnlichkeitsrechte unberhrt.
# 
# Weitere Details zur Lizenzbestimmung gibt es hier:
#   Kurzform: http://creativecommons.org/licenses/by-nc-sa/3.0/de/
#   Komplett: http://creativecommons.org/licenses/by-nc-sa/3.0/de/legalcode
#
###############################################################################
#
# 2009-03-06 0.0.1 rw created
# 2009-04-01 0.1.0 rw RC1
# 2009-05-13 0.1.1 rw MapGpsAt() added
# 2009-05-22 0.1.2 rw add default Map dir
# 2009-07-22 0.1.3 rw Map Offset_x, Offset_y added
# 2009-08-08 0.1.4 rw atan -> atan2
# 2009-11-04 0.1.5 rw Lon > 180
#
###############################################################################

$Version{'libmap.pl'} = "0.1.4 - 2009-08-08";

use Geo::Ellipsoid;     # http://search.cpan.org/dist/Geo-Ellipsoid-1.12/lib/Geo/Ellipsoid.pm
use Math::Trig;

if ( ! defined Cfg->{'map'}->{'MapDir'} )
    {
    # set default map directory
    $Cfg->{'map'}->{'MapDir'} = "map";
    }
require "$Cfg->{'map'}->{'MapDir'}/map.pl";   # Landkarte

#
# Convert GPS (Lat,Lon) to pixel coordinates in map
#
sub MapGps2XY ()
    {
    my ($Lat, $Lon, $Bearing) = @_;
	
    # Aktuell gltige Karte
    my %Map = %{$Maps{'Current'}};

    my $Map_Geo = Geo::Ellipsoid->new( 'units' => 'degrees',
                                       'distance_units' => 'meter',
                                       'ellipsoid' => 'WGS84',
                                       'longitude' => 1,         # Symmetric: -pi..pi
                                     );
								
    # P1 -> P2: Entfernung und Richtung
    my ($P1P2_Dist_m, $P1P2_Bearing) = $Map_Geo->to( $Map{'P1_Lat'}, $Map{'P1_Lon'}, $Map{'P2_Lat'}, $Map{'P2_Lon'} );

    my $dx_p = $Map{'P2_x'} - $Map{'P1_x'};
    my $dy_p = $Map{'P2_y'} - $Map{'P1_y'};
    my $P1P2_Dist_p = sqrt($dx_p*$dx_p + $dy_p*$dy_p);
    my $PixRes = $P1P2_Dist_m / $P1P2_Dist_p ;        # 1 Pixel = $PixRes Meter

    my $Phi = rad2deg atan2 ($dy_p, $dx_p);
    my $PhiRef = $P1P2_Bearing - $Phi;                # Winkel zwischen N und Bild-Horizont

    # P1 -> Target: Entfernung und Richtung
    my ($T_Dist_m, $T_Bearing) = $Map_Geo->to( $Map{'P1_Lat'}, $Map{'P1_Lon'}, $Lat, $Lon );

    my $Alpha = deg2rad($T_Bearing - $PhiRef);        # Winkel zw. Bild-Horizont und Target
    my $T_dx_p = cos($Alpha) * $T_Dist_m / $PixRes;
    my $T_dy_p = sin($Alpha) * $T_Dist_m / $PixRes;

    my $X_p = $Map{'P1_x'} + $T_dx_p + $Map{'Offset_x'};
    my $Y_p = $Map{'P1_y'} + $T_dy_p - $Map{'Offset_y'};

    # map calibration adjustment
    $X_p += $Map{'Offset_x'};
    $Y_p -= $Map{'Offset_y'};   
	
    $Angel = $Bearing - $PhiRef;
	
    return ($X_p, $Y_p, $Angel);
    }

#
# Convert pixel coordinates in map to GPS (Lat,Lon)
#
sub MapXY2Gps ()
    {
    my ($X, $Y) = @_;

    # Aktuell gltige Karte
    my %Map = %{$Maps{'Current'}};

    # map calibration adjustment
    $X -= $Map{'Offset_x'};
    $Y += $Map{'Offset_y'};
	
    my $Map_Geo = Geo::Ellipsoid->new( 'units' => 'degrees',
                                       'distance_units' => 'meter',
                                       'ellipsoid' => 'WGS84',
                                       'longitude' => 1,         # Symmetric: -pi..pi
                                     );
								
    # P1 -> P2: Entfernung und Richtung
    my ($P1P2_Dist_m, $P1P2_Bearing) = $Map_Geo->to( $Map{'P1_Lat'}, $Map{'P1_Lon'}, $Map{'P2_Lat'}, $Map{'P2_Lon'} );

    my $dx_p = $Map{'P2_x'} - $Map{'P1_x'};
    my $dy_p = $Map{'P2_y'} - $Map{'P1_y'};
    my $P1P2_Dist_p = sqrt($dx_p*$dx_p + $dy_p*$dy_p);
    my $PixRes = $P1P2_Dist_m / $P1P2_Dist_p ;        # 1 Pixel = $PixRes Meter

    my $Phi = rad2deg atan2 ($dy_p, $dx_p);
    my $PhiRef = $P1P2_Bearing - $Phi;                # Winkel zwischen N und Bild-Horizont

    my $dx = $X - $Map{'P1_x'};
    my $dy = $Y - $Map{'P1_y'};

    my $Phi = rad2deg atan2 ($dy, $dx);
    my $Bearing = $PhiRef + $Phi;
    my $Dist = $PixRes * sqrt($dx*$dx + $dy*$dy);

    my ($Lat, $Lon) = $Map_Geo->at( $Map{'P1_Lat'}, $Map{'P1_Lon'}, $Dist, $Bearing );

    return ($Lat, $Lon);
    }

#
# Get Bearing, Distance from 2 GPS Points
#
sub MapGpsTo()
    {
    my ($Lat1, $Lon1, $Lat2, $Lon2) = @_;
	
    my $Map_Geo = Geo::Ellipsoid->new( 'units' => 'degrees',
                                       'distance_units' => 'meter',
                                       'ellipsoid' => 'WGS84',
                                       'longitude' => 1,         # Symmetric: -pi..pi
                                     );
    my ($Dist, $Bearing) = $Map_Geo->to( $Lat1, $Lon1, $Lat2, $Lon2);
	
    return ($Dist, $Bearing);
    }

#
# Get Target from GPS-Point, Bearing, Distance
#
sub MapGpsAt()
    {
    my ($Lat, $Lon, $Dist, $Bearing) = @_;
	
    my $Map_Geo = Geo::Ellipsoid->new( 'units' => 'degrees',
                                       'distance_units' => 'meter',
                                       'ellipsoid' => 'WGS84',
                                       'longitude' => 1,         # Symmetric: -pi..pi
                                     );
    my ($Target_Lat, $Target_Lon) = $Map_Geo->at( $Lat, $Lon, $Dist, $Bearing);
	
    return ($Target_Lat, $Target_Lon);
    }

	
# Angel geographic North to Map Horizont
sub MapAngel()
    {
    # Aktuell gltige Karte
    my %Map = %{$Maps{'Current'}};
	
    my $Map_Geo = Geo::Ellipsoid->new( 'units' => 'degrees',
                                       'distance_units' => 'meter',
                                       'ellipsoid' => 'WGS84',
                                       'longitude' => 1,         # Symmetric: -pi..pi
                                     );
								
    # P1 -> P2: Entfernung und Richtung
    my ($P1P2_Dist_m, $P1P2_Bearing) = $Map_Geo->to( $Map{'P1_Lat'}, $Map{'P1_Lon'}, $Map{'P2_Lat'}, $Map{'P2_Lon'} );

    my $dx_p = $Map{'P2_x'} - $Map{'P1_x'};
    my $dy_p = $Map{'P2_y'} - $Map{'P1_y'};

    my $Phi = rad2deg atan2 ($dy_p, $dx_p);
    my $PhiRef = $P1P2_Bearing - $Phi;                # Winkel zwischen N und Bild-Horizont	
	
    return ($PhiRef);
    }
	
1;

__END__
